# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python interface to Hanabi code."""
import os
import re
import cffi
import enum
import sys

DEFAULT_CDEF_PREFIXES = (None, ".", os.path.dirname(__file__), "/include")
DEFAULT_LIB_PREFIXES = (None, ".", os.path.dirname(__file__), "/lib")
PYHANABI_HEADER = "pyhanabi.h"
PYHANABI_LIB = ["libpyhanabi.so", "libpyhanabi.dylib"]
COLOR_CHAR = ["R", "Y", "G", "W", "B"]  # consistent with hanabi_lib/util.cc
CHANCE_PLAYER_ID = -1

ffi = cffi.FFI()
lib = None
cdef_loaded_flag = False
lib_loaded_flag = False


if sys.version_info < (3,):
  def encode_ffi_string(x):
    return ffi.string(x)
else:
  def encode_ffi_string(x):
    return str(ffi.string(x), 'ascii')

def try_cdef(header=PYHANABI_HEADER, prefixes=DEFAULT_CDEF_PREFIXES):
  """Try parsing library header file. Must be called before any pyhanabi calls.

  Args:
    header: filename of pyhanabi header file.
    prefixes: list of paths to search for pyhanabi header file.
  Returns:
    True if header was successfully parsed, False on failure.
  """
  global cdef_loaded_flag
  if cdef_loaded_flag: return True
  for prefix in prefixes:
    try:
      cdef_file = header if prefix is None else prefix + "/" + header
      lines = open(cdef_file).readlines()
      reading_cdef = False
      cdef_string = ""
      for line in lines:
        line = line.rstrip()
        if re.match("extern *\"C\" *{", line):
          reading_cdef = True
          continue
        elif re.match("} */[*] *extern *\"C\" *[*]/", line):
          reading_cdef = False
          continue
        if reading_cdef:
          cdef_string = cdef_string + line + "\n"
      ffi.cdef(cdef_string)
      cdef_loaded_flag = True
      return True
    except IOError:
      pass
  return False


def try_load(library=None, prefixes=DEFAULT_LIB_PREFIXES):
  """Try loading library. Must be called before any pyhanabi calls.

  Args:
    library: filename of pyhanabi library file.
    prefixes: list of paths to search for pyhanabi library file.
  Returns:
    True if library was successfully loaded, False on failure.
  """
  global lib_loaded_flag
  global lib
  if lib_loaded_flag: return True
  if library is None:
    libnames = PYHANABI_LIB
  elif type(library) in (list, tuple):
    libnames = library
  else:
    libnames = (library,)
  for prefix in prefixes:
    for libname in libnames:
      try:
        lib_file = libname if prefix is None else prefix + "/" + libname
        lib = ffi.dlopen(lib_file)
        lib_loaded_flag = True
        return True
      except OSError:
        pass
  return False


def cdef_loaded():
  """Return True if pyhanabi header has been successfully parsed."""
  return cdef_loaded_flag


def lib_loaded():
  """Return True if pyhanabi library has been successfully loaded."""
  return lib_loaded_flag


def color_idx_to_char(color_idx):
  """Helper function for converting color index to a character.

  Args:
    color_idx: int, index into color char vector.

  Returns:
    color_char: str, A single character representing a color.

  Raises:
    AssertionError: If index is not in range.
  """
  assert isinstance(color_idx, int)
  if color_idx == -1:
    return None
  else:
    return COLOR_CHAR[color_idx]


def color_char_to_idx(color_char):
  r"""Helper function for converting color character to index.

  Args:
    color_char: str, Character representing a color.

  Returns:
    color_idx: int, Index into a color array \in [0, num_colors -1]

  Raises:
    ValueError: If color_char is not a valid color.
  """
  assert isinstance(color_char, str)
  try:
    return next(idx for (idx, c) in enumerate(COLOR_CHAR) if c == color_char)
  except StopIteration:
    raise ValueError("Invalid color: {}. Should be one of {}.".format(
        color_char, COLOR_CHAR))


class HanabiCard(object):
  """Hanabi card, with a color and a rank.

  Python implementation of C++ HanabiCard class.
  """

  def __init__(self, color, rank):
    """A simple HanabiCard object.

    Args:
      color: an integer, starting at 0. Colors are in this order RYGWB.
      rank: an integer, starting at 0 (representing a 1 card). In the standard
          game, the largest value is 4 (representing a 5 card).
    """
    self._color = color
    self._rank = rank

  def color(self):
    return self._color

  def rank(self):
    return self._rank

  def __str__(self):
    if self.valid():
      return COLOR_CHAR[self._color] + str(self._rank + 1)
    else:
      return "XX"

  def __repr__(self):
    return self.__str__()

  def __eq__(self, other):
    return self._color == other.color() and self._rank == other.rank()

  def valid(self):
    return self._color >= 0 and self._rank >= 0

  def to_dict(self):
    """Serialize to dict.

    Returns:
      d: dict, containing color and rank of card.
    """
    return {"color": color_idx_to_char(self.color()), "rank": self.rank()}


class HanabiCardKnowledge(object):
  """Accumulated knowledge about color and rank of an initially unknown card.

  Stores two types of knowledge: direct hints about a card, and indirect
  knowledge from hints about other cards.

  For example, say we had two cards that we know nothing about, but our
  partners know are a R1 and B2. Before any hints, both color() and rank()
  return None, and color_plausible(c) and rank_plausible(r) returns True for
  all colors c and ranks r, for both cards.

  Say our partner reveals that our first card is a 1 -- rank index 0. Now for
  the first card we have rank()=0, rank_plausible(0)=True, and
  rank_plausible(r)=False for r != 0. The same hint also tells us the second
  card is NOT a 1 (rank index 0). For the second card, we have rank()=None,
  rank_plausible(0)=False, and rank_plausible(r)=True for r!= 0.

  Note that color() and rank() only consider directly revealed information.
  Both methods will always return None unless the color or rank, respectively,
  are directly revealed. That is, even we have seen hints for all ranks except
  rank index 0, so that rank_plausible(0)=True and rank_plausible(r)=False
  for all r != 0, rank() will still be None.

  Python wrapper of C++ HanabiHand::CardKnowledge class.
  """

  def __init__(self, knowledge):
    self._knowledge = knowledge

  def color(self):
    """Returns color index if exact color was revealed, or None otherwise.

    Does not perform inference to deduce the color from other color hints.
    """
    if lib.ColorWasHinted(self._knowledge):
      return lib.KnownColor(self._knowledge)
    else:
      return None

  def color_plausible(self, color_index):
    """Returns true if we have no hint saying card is not the given color.

    Args:
      color_index: 0-based color index.
    """
    return lib.ColorIsPlausible(self._knowledge, color_index)

  def rank(self):
    """Returns rank index if exact rank was revealed, or None otherwise.

    Does not perform inference to deduce the rank from other rank hints.
    """
    if lib.RankWasHinted(self._knowledge):
      return lib.KnownRank(self._knowledge)
    else:
      return None

  def rank_plausible(self, rank_index):
    """Returns true if we have no hint saying card is not the given rank.

    Args:
      rank_index: 0-based rank index.
    """
    return lib.RankIsPlausible(self._knowledge, rank_index)

  def __str__(self):
    c_string = lib.CardKnowledgeToString(self._knowledge)
    string = encode_ffi_string(c_string)
    lib.DeleteString(c_string)
    return string

  def __repr__(self):
    return self.__str__()

  def to_dict(self):
    """Serialize to dict.

    Returns:
      d: dict, containing color and rank of hint.
    """
    return {"color": color_idx_to_char(self.color()), "rank": self.rank()}


class HanabiMoveType(enum.IntEnum):
  """Move types, consistent with hanabi_lib/hanabi_move.h."""
  INVALID = 0
  PLAY = 1
  DISCARD = 2
  REVEAL_COLOR = 3
  REVEAL_RANK = 4
  DEAL = 5


class HanabiMove(object):
  """Description of an agent move or chance event.

  Python wrapper of C++ HanabiMove class.
  """

  def __init__(self, move):
    assert move is not None
    self._move = move

  @property
  def c_move(self):
    return self._move

  def type(self):
    return HanabiMoveType(lib.MoveType(self._move))

  def card_index(self):
    """Returns 0-based card index for PLAY and DISCARD moves."""
    return lib.CardIndex(self._move)

  def target_offset(self):
    """Returns target player offset for REVEAL_XYZ moves."""
    return lib.TargetOffset(self._move)

  def color(self):
    """Returns 0-based color index for REVEAL_COLOR and DEAL moves."""
    return lib.MoveColor(self._move)

  def rank(self):
    """Returns 0-based rank index for REVEAL_RANK and DEAL moves."""
    return lib.MoveRank(self._move)

  @staticmethod
  def get_discard_move(card_index):
    c_move = ffi.new("pyhanabi_move_t*")
    assert lib.GetDiscardMove(card_index, c_move)
    return HanabiMove(c_move)

  @staticmethod
  def get_play_move(card_index):
    c_move = ffi.new("pyhanabi_move_t*")
    assert lib.GetPlayMove(card_index, c_move)
    return HanabiMove(c_move)

  @staticmethod
  def get_reveal_color_move(target_offset, color):
    """current player is 0, next player clockwise is target_offset 1, etc."""
    c_move = ffi.new("pyhanabi_move_t*")
    assert lib.GetRevealColorMove(target_offset, color, c_move)
    return HanabiMove(c_move)

  @staticmethod
  def get_reveal_rank_move(target_offset, rank):
    """current player is 0, next player clockwise is target_offset 1, etc."""
    c_move = ffi.new("pyhanabi_move_t*")
    assert lib.GetRevealRankMove(target_offset, rank, c_move)
    return HanabiMove(c_move)

  def __str__(self):
    c_string = lib.MoveToString(self._move)
    string = encode_ffi_string(c_string)
    lib.DeleteString(c_string)
    return string

  def __repr__(self):
    return self.__str__()

  def __del__(self):
    if self._move is not None:
      lib.DeleteMove(self._move)
      self._move = None
    del self

  def to_dict(self):
    """Serialize to dict.

    Returns:
      d: dict, Containing type and information of a hanabi move.

    Raises:
      ValueError: If move type is not supported.
    """
    move_dict = {}
    move_type = self.type()
    move_dict["action_type"] = move_type.name
    if move_type == HanabiMoveType.PLAY or move_type == HanabiMoveType.DISCARD:
      move_dict["card_index"] = self.card_index()
    elif move_type == HanabiMoveType.REVEAL_COLOR:
      move_dict["target_offset"] = self.target_offset()
      move_dict["color"] = color_idx_to_char(self.color())
    elif move_type == HanabiMoveType.REVEAL_RANK:
      move_dict["target_offset"] = self.target_offset()
      move_dict["rank"] = self.rank()
    elif move_type == HanabiMoveType.DEAL:
      move_dict["color"] = color_idx_to_char(self.color())
      move_dict["rank"] = self.rank()
    else:
      raise ValueError("Unsupported move: {}".format(self))

    return move_dict


class HanabiHistoryItem(object):
  """A move that has been made within a game, along with the side-effects.

  For example, a play move simply selects a card index between 0-5, but after
  making the move, there is an associated color and rank for the selected card,
  a possibility that the card was successfully added to the fireworks, and an
  information token added if the firework stack was completed.

  Python wrapper of C++ HanabiHistoryItem class.
  """

  def __init__(self, item):
    self._item = item

  def move(self):
    c_move = ffi.new("pyhanabi_move_t*")
    lib.HistoryItemMove(self._item, c_move)
    return HanabiMove(c_move)

  def player(self):
    return lib.HistoryItemPlayer(self._item)

  def scored(self):
    """Play move succeeded in placing card on fireworks."""
    return bool(lib.HistoryItemScored(self._item))

  def information_token(self):
    """Play/Discard move increased the number of information tokens."""
    return bool(lib.HistoryItemInformationToken(self._item))

  def color(self):
    """Color index of card that was Played/Discarded."""
    return lib.HistoryItemColor(self._item)

  def rank(self):
    """Rank index of card that was Played/Discarded."""
    return lib.HistoryItemRank(self._item)

  def card_info_revealed(self):
    """Returns information about whether color/rank was revealed.

    Indices where card i color/rank matches the reveal move. E.g.,
    for Reveal player 1 color red when player 1 has R1 W1 R2 R4 __ the
    result would be [0, 2, 3].
    """
    revealed = []
    bitmask = lib.HistoryItemRevealBitmask(self._item)
    for i in range(8):  # 8 bits in reveal_bitmask
      if bitmask & (1 << i):
        revealed.append(i)
    return revealed

  def card_info_newly_revealed(self):
    """Returns information about whether color/rank was newly revealed.

    Indices where card i color/rank was not previously known. E.g.,
    for Reveal player 1 color red when player 1 has R1 W1 R2 R4 __ the
    result might be [2, 3].  Cards 2 and 3 were revealed to be red,
    but card 0 was previously known to be red, so nothing new was
    revealed. Card 4 is missing, so nothing was revealed about it.
    """
    revealed = []
    bitmask = lib.HistoryItemNewlyRevealedBitmask(self._item)
    for i in range(8):  # 8 bits in reveal_bitmask
      if bitmask & (1 << i):
        revealed.append(i)
    return revealed

  def deal_to_player(self):
    """player that card was dealt to for Deal moves."""
    return lib.HistoryItemDealToPlayer(self._item)

  def __str__(self):
    c_string = lib.HistoryItemToString(self._item)
    string = encode_ffi_string(c_string)
    lib.DeleteString(c_string)
    return string

  def __repr__(self):
    return self.__str__()

  def __del__(self):
    if self._item is not None:
      lib.DeleteHistoryItem(self._item)
      self._item = None
    del self


class HanabiEndOfGameType(enum.IntEnum):
  """Possible end-of-game conditions, consistent with hanabi_state.h."""
  NOT_FINISHED = 0
  OUT_OF_LIFE_TOKENS = 1
  OUT_OF_CARDS = 2
  COMPLETED_FIREWORKS = 3


class HanabiState(object):
  """Current environment state for an active Hanabi game.

  The game is turn-based, with only one active agent at a time. Chance events
  are explicitly included, so the active agent may be "nature" (represented
  by cur_player() returning CHANCE_PLAYER_ID).

  Python wrapper of C++ HanabiState class.
  """

  def __init__(self, game, c_state=None):
    """Returns a new state.

    Args:
      game: HanabiGame describing the parameters for a game of Hanabi.
      c_state: C++ state to copy, or None for a new state.

    NOTE: If c_state is supplied, game is ignored and c_state game is used.
    """
    self._state = ffi.new("pyhanabi_state_t*")
    if c_state is None:
      self._game = game.c_game
      lib.NewState(self._game, self._state)
    else:
      self._game = lib.StateParentGame(c_state)
      lib.CopyState(c_state, self._state)

  def copy(self):
    """Returns a copy of the state."""
    return HanabiState(None, self._state)

  def observation(self, player):
    """Returns player's observed view of current environment state."""
    return HanabiObservation(self._state, self._game, player)

  def apply_move(self, move):
    """Advance the environment state by making move for acting player."""
    lib.StateApplyMove(self._state, move.c_move)

  def cur_player(self):
    """Returns index of next player to act.

    Index will be CHANCE_PLAYER_ID if a chance event needs to be resolved.
    """
    return lib.StateCurPlayer(self._state)

  def deck_size(self):
    """Returns number of cards left in the deck."""
    return lib.StateDeckSize(self._state)

  def discard_pile(self):
    """Returns a list of all discarded cards, in order they were discarded."""
    discards = []
    c_card = ffi.new("pyhanabi_card_t*")
    for index in range(lib.StateDiscardPileSize(self._state)):
      lib.StateGetDiscard(self._state, index, c_card)
      discards.append(HanabiCard(c_card.color, c_card.rank))
    return discards

  def fireworks(self):
    """Returns a list of fireworks levels by value, ordered by color (RYGWB).

    Important note on representation / format: when no fireworks have been
    played, this function returns [0, 0, 0, 0, 0]. When only the red 1 has been
    played, this function returns [1, 0, 0, 0, 0].
    """
    firework_list = []
    num_colors = lib.NumColors(self._game)
    for c in range(num_colors):
      firework_list.append(lib.StateFireworks(self._state, c))
    return firework_list

  def deal_random_card(self):
    """If cur_player == CHANCE_PLAYER_ID, make a random card-deal move."""
    lib.StateDealRandomCard(self._state)

  def player_hands(self):
    """Returns a list of all hands, with cards ordered oldest to newest."""
    hand_list = []
    c_card = ffi.new("pyhanabi_card_t*")
    for pid in range(self.num_players()):
      player_hand = []
      hand_size = lib.StateGetHandSize(self._state, pid)
      for i in range(hand_size):
        lib.StateGetHandCard(self._state, pid, i, c_card)
        player_hand.append(HanabiCard(c_card.color, c_card.rank))
      hand_list.append(player_hand)
    return hand_list

  def information_tokens(self):
    """Returns the number of information tokens remaining."""
    return lib.StateInformationTokens(self._state)

  def end_of_game_status(self):
    """Returns the end of game status, NOT_FINISHED if game is still active."""
    return HanabiEndOfGameType(lib.StateEndOfGameStatus(self._state))

  def is_terminal(self):
    """Returns false if game is still active, true otherwise."""
    return (lib.StateEndOfGameStatus(self._state) !=
            HanabiEndOfGameType.NOT_FINISHED)

  def legal_moves(self):
    """Returns list of legal moves for currently acting player."""
    moves = []
    c_movelist = lib.StateLegalMoves(self._state)
    num_moves = lib.NumMoves(c_movelist)
    for i in range(num_moves):
      c_move = ffi.new("pyhanabi_move_t*")
      lib.GetMove(c_movelist, i, c_move)
      moves.append(HanabiMove(c_move))
    lib.DeleteMoveList(c_movelist)
    return moves

  def move_is_legal(self, move):
    """Returns true if and only if move is legal for active agent."""
    return lib.MoveIsLegal(self._state, move.c_move)

  def card_playable_on_fireworks(self, color, rank):
    """Returns true if and only if card can be successfully played.

    Args:
      color: 0-based color index of card
      rank: 0-based rank index of card
    """
    return lib.CardPlayableOnFireworks(self._state, color, rank)

  def life_tokens(self):
    """Returns the number of information tokens remaining."""
    return lib.StateLifeTokens(self._state)

  def num_players(self):
    """Returns the number of players in the game."""
    return lib.StateNumPlayers(self._state)

  def score(self):
    """Returns the co-operative game score at a terminal state.

    NOTE: result is undefined when game is NOT_FINISHED.
    """
    return lib.StateScore(self._state)

  def move_history(self):
    """Returns list of moves made, from oldest to most recent."""
    history = []
    history_len = lib.StateLenMoveHistory(self._state)
    for i in range(history_len):
      c_history_item = ffi.new("pyhanabi_history_item_t*")
      lib.StateGetMoveHistory(self._state, i, c_history_item)
      history.append(HanabiHistoryItem(c_history_item))
    return history

  def __str__(self):
    c_string = lib.StateToString(self._state)
    string = encode_ffi_string(c_string)
    lib.DeleteString(c_string)
    return string

  def __repr__(self):
    return self.__str__()

  def __del__(self):
    if self._state is not None:
      lib.DeleteState(self._state)
      self._state = None
    del self


class AgentObservationType(enum.IntEnum):
  """Possible agent observation types, consistent with hanabi_game.h.

  A kMinimal observation is similar to what a human sees, and does not
  include any memory of past RevalColor/RevealRank hints. A CardKnowledge
  observation includes per-card knowledge of past hints, as well as simple
  inferred knowledge of the form "this card is not red, because it was
  not revealed as red in a past <RevealColor Red> move. A Seer observation
  shows all cards, including the player's own cards, regardless of what
  hints have been given.
  """
  MINIMAL = 0
  CARD_KNOWLEDGE = 1
  SEER = 2


class HanabiGame(object):
  """Game parameters describing a specific instance of Hanabi.

  Python wrapper of C++ HanabiGame class.
  """

  def __init__(self, params=None):
    """Creates a HanabiGame object.

    Args:
      params: is a dictionary of parameters and their values.

    Possible parameters include
    "players": 2 <= number of players <= 5
    "colors": 1 <= number of different card colors in deck <= 5
    "rank": 1 <= number of different card ranks in deck <= 5
    "hand_size": 1 <= number of cards in player hand
    "max_information_tokens": 1 <= maximum (and initial) number of info tokens.
    "max_life_tokens": 1 <= maximum (and initial) number of life tokens.
    "seed": random number seed. -1 to use system random device to get seed.
    "random_start_player": boolean. If true, start with random player, not 0.
    "observation_type": int AgentObservationType.
    """
    if params is None:
      self._game = ffi.new("pyhanabi_game_t*")
      lib.NewDefaultGame(self._game)
    else:
      param_list = []
      for key in params:
        param_list.append(ffi.new("char[]", key.encode('ascii')))
        param_list.append(ffi.new("char[]", str(params[key]).encode('ascii')))
      c_array = ffi.new("char * [" + str(len(param_list)) + "]", param_list)
      self._game = ffi.new("pyhanabi_game_t*")
      lib.NewGame(self._game, len(param_list), c_array)

  def new_initial_state(self):
    return HanabiState(self)

  @property
  def c_game(self):
    """Return the C++ HanabiGame object."""
    return self._game

  def __del__(self):
    if self._game is not None:
      lib.DeleteGame(self._game)
      self._game = None
    del self

  def parameter_string(self):
    """Returns string with all parameter choices."""
    c_string = lib.GameParamString(self._game)
    string = encode_ffi_string(c_string)
    lib.DeleteString(c_string)
    return string

  def num_players(self):
    """Returns the number of players in the game."""
    return lib.NumPlayers(self._game)

  def num_colors(self):
    """Returns the number of card colors in the initial deck."""
    return lib.NumColors(self._game)

  def num_ranks(self):
    """Returns the number of card ranks in the initial deck."""
    return lib.NumRanks(self._game)

  def hand_size(self):
    """Returns the maximum number of cards in each player hand.

    The number of cards in a player's hand may be smaller than this maximum
    a) at the beginning of the game before cards are dealt out, b) after
    any Play or Discard action and before the subsequent deal event, and c)
    after the deck is empty and cards can no longer be dealt to a player.
    """
    return lib.HandSize(self._game)

  def max_information_tokens(self):
    """Returns the initial number of information tokens."""
    return lib.MaxInformationTokens(self._game)

  def max_life_tokens(self):
    """Returns the initial number of life tokens."""
    return lib.MaxLifeTokens(self._game)

  def observation_type(self):
    return AgentObservationType(lib.ObservationType(self._game))

  def max_moves(self):
    """Returns the number of possible legal moves in the game."""
    return lib.MaxMoves(self._game)

  def num_cards(self, color, rank):
    """Returns number of instances of Card(color, rank) in the initial deck."""
    return lib.NumCards(self._game, color, rank)

  def get_move_uid(self, move):
    """Returns a unique ID describing a legal move, or -1 for invalid move."""
    return lib.GetMoveUid(self._game, move.c_move)

  def get_move(self, move_uid):
    """Returns a HanabiMove represented by 0 <= move_uid < max_moves()."""
    move = ffi.new("pyhanabi_move_t*")
    lib.GetMoveByUid(self._game, move_uid, move)
    return HanabiMove(move)


class HanabiObservation(object):
  """Player's observed view of an environment HanabiState.

  The main differences are that 1) a player's own cards are not visible, and
  2) a player does not know their own player index (seat) so that all player
  indices are described relative to the observing player (or equivalently,
  that from the player's point of view, they are always player index 0).

  Python wrapper of C++ HanabiObservation class.
  """

  def __init__(self, state, game, player):
    """Construct using HanabiState.observation(player)."""
    self._observation = ffi.new("pyhanabi_observation_t*")
    self._game = game
    lib.NewObservation(state, player, self._observation)

  def __str__(self):
    c_string = lib.ObsToString(self._observation)
    string = encode_ffi_string(c_string)
    lib.DeleteString(c_string)
    return string

  def __repr__(self):
    return self.__str__()

  def __del__(self):
    if self._observation is not None:
      lib.DeleteObservation(self._observation)
      self._observation = None
    del self

  def observation(self):
    """Returns the C++ HanabiObservation object."""
    return self._observation

  def cur_player_offset(self):
    """Returns the player index of the acting player, relative to observer."""
    return lib.ObsCurPlayerOffset(self._observation)

  def num_players(self):
    """Returns the number of players in the game."""
    return lib.ObsNumPlayers(self._observation)

  def observed_hands(self):
    """Returns a list of all hands, with cards ordered oldest to newest.

     The observing player's cards are always invalid.
    """
    hand_list = []
    c_card = ffi.new("pyhanabi_card_t*")
    for pid in range(self.num_players()):
      player_hand = []
      hand_size = lib.ObsGetHandSize(self._observation, pid)
      for i in range(hand_size):
        lib.ObsGetHandCard(self._observation, pid, i, c_card)
        player_hand.append(HanabiCard(c_card.color, c_card.rank))
      hand_list.append(player_hand)
    return hand_list

  def card_knowledge(self):
    """Returns a per-player list of hinted card knowledge.

    Each player's entry is a per-card list of HanabiCardKnowledge objects.
    Each HanabiCardKnowledge for a card gives the knowledge about the cards
    accumulated over all past reveal actions.
    """
    card_knowledge_list = []
    for pid in range(self.num_players()):
      player_card_knowledge = []
      hand_size = lib.ObsGetHandSize(self._observation, pid)
      for i in range(hand_size):
        c_knowledge = ffi.new("pyhanabi_card_knowledge_t*")
        lib.ObsGetHandCardKnowledge(self._observation, pid, i, c_knowledge)
        player_card_knowledge.append(HanabiCardKnowledge(c_knowledge))
      card_knowledge_list.append(player_card_knowledge)
    return card_knowledge_list

  def discard_pile(self):
    """Returns a list of all discarded cards, in order they were discarded."""
    discards = []
    c_card = ffi.new("pyhanabi_card_t*")
    for index in range(lib.ObsDiscardPileSize(self._observation)):
      lib.ObsGetDiscard(self._observation, index, c_card)
      discards.append(HanabiCard(c_card.color, c_card.rank))
    return discards

  def fireworks(self):
    """Returns a list of fireworks levels by value, ordered by color."""
    firework_list = []
    num_colors = lib.NumColors(self._game)
    for c in range(num_colors):
      firework_list.append(lib.ObsFireworks(self._observation, c))
    return firework_list

  def deck_size(self):
    """Returns number of cards left in the deck."""
    return lib.ObsDeckSize(self._observation)

  def last_moves(self):
    """Returns moves made since observing player last acted.

    Each entry in list is a HanabiHistoryItem, ordered from most recent
    move to oldest.  Oldest move is the last action made by observing
    player. Skips initial chance moves to deal hands.
    """
    history_items = []
    for i in range(lib.ObsNumLastMoves(self._observation)):
      history_item = ffi.new("pyhanabi_history_item_t*")
      lib.ObsGetLastMove(self._observation, i, history_item)
      history_items.append(HanabiHistoryItem(history_item))
    return history_items

  def information_tokens(self):
    """Returns the number of information tokens remaining."""
    return lib.ObsInformationTokens(self._observation)

  def life_tokens(self):
    """Returns the number of information tokens remaining."""
    return lib.ObsLifeTokens(self._observation)

  def legal_moves(self):
    """Returns list of legal moves for observing player.

    List is empty if cur_player() != 0 (observer is not currently acting).
    """
    moves = []
    for i in range(lib.ObsNumLegalMoves(self._observation)):
      move = ffi.new("pyhanabi_move_t*")
      lib.ObsGetLegalMove(self._observation, i, move)
      moves.append(HanabiMove(move))
    return moves

  def card_playable_on_fireworks(self, color, rank):
    """Returns true if and only if card can be successfully played.

    Args:
      color: 0-based color index of card
      rank: 0-based rank index of card
    """
    return lib.ObsCardPlayableOnFireworks(self._observation, color, rank)


class ObservationEncoderType(enum.IntEnum):
  """Encoder types, consistent with observation_encoder.h."""
  CANONICAL = 0


class ObservationEncoder(object):
  """ObservationEncoder class.

  The canonical observations wrap an underlying C++ class. To make custom
  observation encoders, create a subclass of this base class and override
  the shape and encode methods.
  """

  def __init__(self, game, enc_type=ObservationEncoderType.CANONICAL):
    """Construct using HanabiState.observation(player)."""
    self._game = game.c_game
    self._encoder = ffi.new("pyhanabi_observation_encoder_t*")
    lib.NewObservationEncoder(self._encoder, self._game, enc_type)

  def __del__(self):
    if self._encoder is not None:
      lib.DeleteObservationEncoder(self._encoder)
      self._encoder = None
      self._game = None
    del self

  def shape(self):
    c_shape_str = lib.ObservationShape(self._encoder)
    shape_string = encode_ffi_string(c_shape_str)
    lib.DeleteString(c_shape_str)
    shape = [int(x) for x in shape_string.split(",")]
    return shape

  def encode(self, observation):
    """Encode the observation as a sequence of bits."""
    c_encoding_str = lib.EncodeObservation(self._encoder,
                                           observation.observation())
    encoding_string = encode_ffi_string(c_encoding_str)
    lib.DeleteString(c_encoding_str)
    # Canonical observations are bit strings, so it is ok to encode using a
    # string. For float or double observations, make a custom object
    encoding = [int(x) for x in encoding_string.split(",")]
    return encoding


try_cdef()
if cdef_loaded():
  try_load()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plays a round of Tarok with actions from user input."""

import pyspiel


def play_tarok_game():
  game = pyspiel.load_game("tarok(players=3)")
  state = game.new_initial_state()
  while not state.is_terminal():
    print_info(game, state)
    state.apply_action(int(input("Enter action: ")))
    print("-" * 70, "\n")
  print(state.current_game_phase())
  print("Players' scores: {}".format(state.rewards()))


def print_info(unused_game, state):
  """Print information about the game state."""
  print("Game phase: {}".format(state.current_game_phase()))
  print("Selected contract: {}".format(state.selected_contract()))
  print("Current player: {}".format(state.current_player()))
  player_cards = state.player_cards(state.current_player())
  action_names = [state.card_action_to_string(a) for a in player_cards]
  print("\nPlayer cards: {}".format(
      list(zip(action_names, player_cards))))

  if state.current_game_phase() == pyspiel.TarokGamePhase.TALON_EXCHANGE:
    print_talon_exchange_info(state)
  elif state.current_game_phase() == pyspiel.TarokGamePhase.TRICKS_PLAYING:
    print_tricks_playing_info(state)
  else:
    print()

  legal_actions = state.legal_actions()
  action_names = [state.action_to_string(a) for a in state.legal_actions()]
  print("Legal actions: {}\n".format(
      list(zip(action_names, legal_actions))))


def print_talon_exchange_info(state):
  talon = [[state.card_action_to_string(x) for x in talon_set]
           for talon_set in state.talon_sets()]
  print("\nTalon: {}\n".format(talon))


def print_tricks_playing_info(state):
  trick_cards = state.trick_cards()
  action_names = [state.card_action_to_string(a) for a in trick_cards]
  print("\nTrick cards: {}\n".format(
      list(zip(action_names, trick_cards))))


if __name__ == "__main__":
  play_tarok_game()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.exploitability_descent."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python.algorithms import exploitability_descent
import pyspiel

# Temporary disabling of v2 behavior until code is updated.
tf.disable_v2_behavior()


class ExploitabilityDescentTest(tf.test.TestCase):

  def test_solver_kuhn_poker_first_10_steps_convergence(self):
    solver = exploitability_descent.Solver(pyspiel.load_game("kuhn_poker"))
    with self.session() as session:
      session.run(tf.global_variables_initializer())
      nash_conv = [solver.step(session, learning_rate=1.0) for _ in range(11)]

    # Nash conv is 2x exploitability. Values taken from test run, not
    # independently verified.
    np.testing.assert_allclose(nash_conv, [
        0.91666666666666652, 0.67893004801213452, 0.48109148836354743,
        0.40061420923255808, 0.36617242161468722, 0.33676996443499557,
        0.30925081512398128, 0.28827843035940964, 0.26830042206858751,
        0.24418597846799289, 0.22168699344791482
    ])

  def test_solver_leduc_poker_first_10_steps_convergence(self):
    solver = exploitability_descent.Solver(pyspiel.load_game("leduc_poker"))
    with self.session() as session:
      session.run(tf.global_variables_initializer())
      nash_conv = [solver.step(session, learning_rate=1.0) for _ in range(11)]

    # Nash conv is 2x exploitability. Values taken from test run, not
    # independently verified.
    np.testing.assert_allclose(nash_conv, [
        4.7472224, 4.3147216, 3.9900389, 3.7576618, 3.5771275, 3.4414644,
        3.3272073, 3.1898201, 3.1089299, 3.0108435, 2.8992782
    ])


if __name__ == "__main__":
  tf.test.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualizing public trees with graphviz.

GamePublicTree builds a `pygraphviz.AGraph` reprensentation of the public tree.
It extends the module `treeviz`, look there for more details on how to install
dependencies and advanced usage.

GamePublicTree allows to draw relationships between public and world trees
as well. This can be useful in visualizing games that implement factored
observations and PublicState API.
"""

from absl import logging
from open_spiel.python.visualizations.treeviz import GameTree

# pylint: disable=g-import-not-at-top
try:
  import pygraphviz
except (ImportError, Exception) as e:
  raise ImportError(
      str(e) + "\nPlease make sure to install the following dependencies:\n"
      "sudo apt-get install graphviz libgraphviz-dev\n"
      "pip install pygraphviz")
# pylint: enable=g-import-not-at-top

_PLAYER_SHAPES = {0: "square", 1: "ellipse"}
_PLAYER_COLORS = {-1: "black", 0: "blue", 1: "red"}
_FONTSIZE = 8
_WIDTH = _HEIGHT = 0.25
_ARROWSIZE = .5
_MARGIN = 0.01


def default_public_state_decorator(public_state):
  """Decorates a public state - node of the public tree.

  This method can be called by a custom decorator to prepopulate the attributes
  dictionary. Then only relevant attributes need to be changed, or added.

  Args:
    public_state: The public_state.

  Returns:
    `dict` with graphviz node style attributes.
  """
  attrs = {
      "label": "",
      "fontsize": _FONTSIZE,
      "width": _WIDTH,
      "height": _HEIGHT,
      "margin": _MARGIN
  }
  if public_state.is_terminal():
    attrs["shape"] = "diamond"
  else:
    attrs["label"] = str(public_state.get_public_observation_history())
    attrs["shape"] = "point"
    attrs["width"] = _WIDTH / 2.
    attrs["height"] = _HEIGHT / 2.

  if public_state.is_root():
    attrs["label"] = public_state.get_public_observation_history()[0]
    attrs["shape"] = "circle"
  return attrs


# The decorator can be "overriden", so pylint: disable=unused-argument
def default_transition_decorator(parent, child, transition):
  """Decorates a public transition - an edge of the public tree.

  This method can be called by a custom decorator to prepopulate the attributes
  dictionary. Then only relevant attributes need to be changed, or added.

  Args:
    parent: The parent public state.
    child: The child public state.
    transition: `string` the selected transition in the parent state.

  Returns:
    `dict` with graphviz node style attributes.
  """
  attrs = {
      "label": " " + transition,
      "fontsize": _FONTSIZE,
      "arrowsize": _ARROWSIZE
  }
  attrs["color"] = "black"
  return attrs


def default_public_to_base_decorator(public_state, state):
  """Decorates an edge going from public state to the public set.

  This method can be called by a custom decorator to prepopulate the attributes
  dictionary. Then only relevant attributes need to be changed, or added.

  Args:
    public_state: The public state.
    state: A state within the public set (i.e. the cluster of states)

  Returns:
    `dict` with graphviz node style attributes.
  """
  attrs = {
      "label": "",
      "fontsize": _FONTSIZE,
      "arrowsize": _ARROWSIZE,
      "penwidth": 7,
      "color": "#88888888",
      "style": "solid",
      "splines": False,
      "constraint": False,
  }
  return attrs


class GamePublicTree(GameTree, pygraphviz.AGraph):
  """Builds `pygraphviz.AGraph` of the game tree.

  Attributes:
    public_game: A `pyspiel.GameWithPublicStates` object.
    draw_world: Should we draw the world tree? This is accomplished by calling
      drawing of treeviz.GameTree
    target_public_to_base: If we draw the world tree, we can draw an arrow that
      connects the public tree and world tree. This is done only in one place,
      because it would clutter the graph.
    depth_limit: Maximum depth of the tree. Optional, default=-1 (no limit).
    public_state_decorator: Decorator function for nodes (public states).
      Optional, default=`public_tree_viz.default_public_state_decorator`.
    transition_decorator: Decorator function for edges (public transitions).
      Optional, default=`public_tree_viz.default_transition_decorator`.
    public_to_base_decorator: Decorator function for arrows (between public tree
      and the world tree). Optional,
      default=`public_tree_viz.default_public_to_base_decorator`.
    kwargs: Keyword arguments passed on to `pygraphviz.AGraph.__init__` and to
      `pyspiel.treeviz.GameTree.__init__`.
  """

  def __init__(self,
               public_game=None,
               draw_world=True,
               target_public_to_base=None,
               depth_limit=-1,
               public_state_decorator=default_public_state_decorator,
               transition_decorator=default_transition_decorator,
               public_to_base_decorator=default_public_to_base_decorator,
               **kwargs):

    kwargs["directed"] = kwargs.get("directed", True)
    if draw_world:
      super(GamePublicTree, self).__init__(
          public_game.get_base_game() if public_game else None,
          depth_limit=depth_limit,
          target_pubset=target_public_to_base,
          **kwargs)
    else:
      super(GamePublicTree, self).__init__()

    # We use pygraphviz.AGraph.add_subgraph to cluster nodes, and it requires a
    # default constructor. Thus game needs to be optional.
    if public_game is None:
      return

    self.public_game = public_game
    self._public_state_decorator = public_state_decorator
    self._transition_decorator = transition_decorator
    self._public_to_base_decorator = public_to_base_decorator

    root_public_state = public_game.new_initial_public_state()
    self.add_node(
        self.public_state_to_str(root_public_state),
        **self._public_state_decorator(root_public_state))
    self._build_public_tree(root_public_state, depth_limit)

    if target_public_to_base:
      found = False
      for pubset, sibblings in self._pubsets.items():
        if target_public_to_base == pubset:
          found = True
          # Let's find that public state
          transitions = pubset.split(",")
          public_state = public_game.new_initial_public_state()
          for transition in transitions[1:]:
            public_state.apply_public_transition(transition)

          state_str = str(public_state.get_public_observation_history())
          state = sibblings[-1]

          self.add_edge(state_str, state,
                        **self._public_to_base_decorator(public_state, state))
      if not found:
        logging.warning(
            "Could not find target public state '%s' "
            "Did you mean one of these?\n%s", target_public_to_base,
            "\n".join(self._pubsets.keys()))

  def public_state_to_str(self, public_state):
    """Unique string representation of a public state.

    Args:
      public_state: The public state.

    Returns:
      String representation of public state.
    """
    return str(public_state.get_public_observation_history())

  def _build_public_tree(self, public_state, depth_limit):
    """Recursively builds the game tree."""
    state_str = self.public_state_to_str(public_state)

    if public_state.is_terminal():
      return
    if public_state.move_number() > depth_limit >= 0:
      return

    for transition in public_state.legal_transitions():
      child = public_state.child(transition)
      child_str = self.public_state_to_str(child)
      self.add_node(child_str, **self._public_state_decorator(child))
      self.add_edge(
          state_str, child_str,
          **self._transition_decorator(public_state, child, transition))

      self._build_public_tree(child, depth_limit)

  def _repr_svg_(self):
    """Allows to render directly in Jupyter notebooks and Google Colab."""
    if not self.has_layout:
      self.layout(prog="dot")
    return self.draw(format="svg").decode(self.encoding)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An example of building and exporting a Tensorflow graph.

Adapted from the Travis Ebesu's blog post:
https://tebesu.github.io/posts/Training-a-TensorFlow-graph-in-C++-API
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
import numpy as np
import tensorflow.compat.v1 as tf
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_string("game", "breakthrough", "Name of the game")
flags.DEFINE_string("dir", "/tmp", "Directory to save graph")
flags.DEFINE_string("filename", "graph.pb", "Filename for the graph")


def main(_):
  game = pyspiel.load_game(FLAGS.game)

  # Information state length
  info_state_shape = game.observation_tensor_shape()
  flat_info_state_length = np.prod(info_state_shape)

  # Output
  num_actions = game.num_distinct_actions()

  with tf.Session() as sess:
    net_input = tf.placeholder(
        tf.float32, [None, flat_info_state_length], name="input")

    # pylint: disable=unused-variable
    output = tf.placeholder(tf.float32, [None, num_actions], name="output")
    legals_mask = tf.placeholder(
        tf.float32, [None, num_actions], name="legals_mask")

    policy_net = tf.layers.dense(net_input, 128, activation=tf.nn.relu)
    policy_net = tf.layers.dense(policy_net, 128, activation=tf.nn.relu)
    policy_net = tf.layers.dense(policy_net, num_actions)

    # Note: subtracting the max here is to help with numerical stability.
    # However, there can still be numerical problems. If you are doing a softmax
    # here, it can return NaN when the max for the policy net is high on one of
    # the illegal actions, because policy_net - max will be small for legal
    # actions, giving all exp(small) == 0 in the denominator, returning NaN at
    # the end. One fix is to set the logits to -inf and define a custom cross
    # entropy op that ignores over the illegal actions.
    policy_net = policy_net - tf.reduce_max(policy_net, axis=-1, keepdims=True)

    masked_exp_logit = tf.multiply(tf.exp(policy_net), legals_mask)
    renormalizing_factor = tf.reduce_sum(
        masked_exp_logit, axis=-1, keepdims=True)
    # pylint: disable=unused-variable
    policy_softmax = tf.where(
        tf.equal(legals_mask, 0.),
        tf.zeros_like(masked_exp_logit),
        tf.divide(masked_exp_logit, renormalizing_factor),
        name="policy_softmax")

    policy_targets = tf.placeholder(shape=[None, num_actions], dtype=tf.float32)

    policy_cost = tf.reduce_mean(
        tf.nn.softmax_cross_entropy_with_logits_v2(
            logits=policy_net, labels=policy_targets),
        axis=0)

    # We make one sample.
    sampled_actions = tf.random.categorical(
        tf.log(policy_softmax), 1, name="sampled_actions")

    # pylint: disable=unused-variable
    optimizer = tf.train.AdamOptimizer(0.0001).minimize(
        policy_cost, name="train")

    # pylint: disable=unused-variable
    init = tf.initialize_variables(tf.all_variables(), name="init_all_vars_op")

    print("Writing file: {}/{}".format(FLAGS.dir, FLAGS.filename))
    tf.train.write_graph(
        sess.graph_def, FLAGS.dir, FLAGS.filename, as_text=False)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Simple AlphaZero tic tac toe example.

Take a look at the log-learner.txt in the output directory.

If you want more control, check out `alpha_zero.py`.
"""

from absl import app
from absl import flags

from open_spiel.python.algorithms.alpha_zero import alpha_zero
from open_spiel.python.utils import spawn

flags.DEFINE_string("path", None, "Where to save checkpoints.")
FLAGS = flags.FLAGS


def main(unused_argv):
  config = alpha_zero.Config(
      game="tic_tac_toe",
      path=FLAGS.path,
      learning_rate=0.01,
      weight_decay=1e-4,
      train_batch_size=128,
      replay_buffer_size=2**14,
      replay_buffer_reuse=4,
      max_steps=25,
      checkpoint_freq=25,

      actors=4,
      evaluators=4,
      uct_c=1,
      max_simulations=20,
      policy_alpha=0.25,
      policy_epsilon=1,
      temperature=1,
      temperature_drop=4,
      evaluation_window=50,
      eval_levels=7,

      nn_model="resnet",
      nn_width=128,
      nn_depth=2,
      observation_shape=None,
      output_size=None,

      quiet=True,
  )
  alpha_zero.alpha_zero(config)


if __name__ == "__main__":
  with spawn.main_handler():
    app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example use of the CFR algorithm on Kuhn Poker."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pickle
from absl import app
from absl import flags

import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_enum("solver", "cfr", ["cfr", "cfrplus", "cfrbr"], "CFR solver")
flags.DEFINE_integer("iterations", 20, "Number of iterations")
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")
flags.DEFINE_integer("players", 2, "Number of players")


def main(_):
  game = pyspiel.load_game(
      FLAGS.game,
      {"players": pyspiel.GameParameter(FLAGS.players)},
  )

  if FLAGS.solver == "cfr":
    solver = pyspiel.CFRSolver(game)
  elif FLAGS.solver == "cfrplus":
    solver = pyspiel.CFRPlusSolver(game)
  elif FLAGS.solver == "cfrbr":
    solver = pyspiel.CFRBRSolver(game)

  for i in range(int(FLAGS.iterations / 2)):
    solver.evaluate_and_update_policy()
    print("Iteration {} exploitability: {:.6f}".format(
        i, pyspiel.exploitability(game, solver.average_policy())))

  print("Persisting the model...")
  with open("{}_solver.pickle".format(FLAGS.solver), "wb") as file:
    pickle.dump(solver, file, pickle.HIGHEST_PROTOCOL)

  print("Loading the model...")
  with open("{}_solver.pickle".format(FLAGS.solver), "rb") as file:
    loaded_solver = pickle.load(file)
  print("Exploitability of the loaded model: {:.6f}".format(
      pyspiel.exploitability(game, loaded_solver.average_policy())))

  for i in range(int(FLAGS.iterations / 2)):
    loaded_solver.evaluate_and_update_policy()
    print("Iteration {} exploitability: {:.6f}".format(
        int(FLAGS.iterations / 2) + i,
        pyspiel.exploitability(game, loaded_solver.average_policy())))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example use of the RCFR algorithm on Kuhn Poker."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
import tensorflow.compat.v1 as tf

from open_spiel.python.algorithms import rcfr
import pyspiel

tf.enable_eager_execution()

FLAGS = flags.FLAGS

flags.DEFINE_integer("iterations", 100, "Number of iterations")
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")
flags.DEFINE_integer("players", 2, "Number of players")
flags.DEFINE_integer("print_freq", 10, "How often to print the exploitability")
flags.DEFINE_boolean("bootstrap", False,
                     "Whether or not to use bootstrap targets")
flags.DEFINE_boolean("truncate_negative", False,
                     "Whether or not to truncate negative targets to zero")
flags.DEFINE_integer(
    "buffer_size", -1,
    "A reservoir buffer size. A non-positive size implies an effectively "
    "infinite buffer.")
flags.DEFINE_integer("num_hidden_layers", 1,
                     "The number of hidden layers in the regret model.")
flags.DEFINE_integer("num_hidden_units", 13,
                     "The number of hidden layers in the regret model.")
flags.DEFINE_integer(
    "num_hidden_factors", 8,
    "The number of factors in each hidden layer in the regret model.")
flags.DEFINE_boolean(
    "use_skip_connections", True,
    "Whether or not to use skip connections in the regret model.")
flags.DEFINE_integer(
    "num_epochs", 200,
    "The number of epochs to run between each iterations to update the regret "
    "models.")
flags.DEFINE_integer("batch_size", 100, "The regret model training batch size.")
flags.DEFINE_float("step_size", 0.01, "The ADAM (AMSGrad) optimizer step size.")


def main(_):
  game = pyspiel.load_game(FLAGS.game,
                           {"players": pyspiel.GameParameter(FLAGS.players)})

  models = []
  for _ in range(game.num_players()):
    models.append(
        rcfr.DeepRcfrModel(
            game,
            num_hidden_layers=FLAGS.num_hidden_layers,
            num_hidden_units=FLAGS.num_hidden_units,
            num_hidden_factors=FLAGS.num_hidden_factors,
            use_skip_connections=FLAGS.use_skip_connections))

  if FLAGS.buffer_size > 0:
    solver = rcfr.ReservoirRcfrSolver(
        game,
        models,
        FLAGS.buffer_size,
        truncate_negative=FLAGS.truncate_negative)
  else:
    solver = rcfr.RcfrSolver(
        game,
        models,
        truncate_negative=FLAGS.truncate_negative,
        bootstrap=FLAGS.bootstrap)

  def _train_fn(model, data):
    """Train `model` on `data`."""
    data = data.shuffle(FLAGS.batch_size * 10)
    data = data.batch(FLAGS.batch_size)
    data = data.repeat(FLAGS.num_epochs)

    optimizer = tf.keras.optimizers.Adam(lr=FLAGS.step_size, amsgrad=True)

    @tf.function
    def _train():
      for x, y in data:
        optimizer.minimize(
            lambda: tf.losses.huber_loss(y, model(x), delta=0.01),  # pylint: disable=cell-var-from-loop
            model.trainable_variables)

    _train()

  # End of _train_fn

  for i in range(FLAGS.iterations):
    solver.evaluate_and_update_policy(_train_fn)
    if i % FLAGS.print_freq == 0:
      conv = pyspiel.exploitability(game, solver.average_policy())
      print("Iteration {} exploitability {}".format(i, conv))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Continuous-time population dynamics."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


def replicator(state, fitness):
  """Continuous-time replicator dynamics.

  This is the standard form of the continuous-time replicator dynamics also
  known as selection dynamics.

  For more details, see equation (5) page 9 in
  https://jair.org/index.php/jair/article/view/10952

  Args:
    state: Probability distribution as an `np.array(shape=num_strategies)`.
    fitness: Fitness vector as an `np.array(shape=num_strategies)`.

  Returns:
    Time derivative of the population state.
  """
  avg_fitness = state.dot(fitness)
  return state * (fitness - avg_fitness)


def boltzmannq(state, fitness, temperature=1.):
  """Selection-mutation dynamics modeling Q-learning with Boltzmann exploration.

  For more details, see equation (10) page 15 in
  https://jair.org/index.php/jair/article/view/10952

  Args:
    state: Probability distribution as an `np.array(shape=num_strategies)`.
    fitness: Fitness vector as an `np.array(shape=num_strategies)`.
    temperature: A scalar parameter determining the rate of exploration.

  Returns:
    Time derivative of the population state.
  """
  exploitation = (1. / temperature) * replicator(state, fitness)
  exploration = (np.log(state) - state.dot(np.log(state).transpose()))
  return exploitation - state * exploration


def qpg(state, fitness):
  """Q-based policy gradient dynamics (QPG).

  For more details, see equation (12) on page 18 in
  https://arxiv.org/pdf/1810.09026.pdf

  Args:
    state: Probability distribution as an `np.array(shape=num_strategies)`.
    fitness: Fitness vector as an `np.array(shape=num_strategies)`.

  Returns:
    Time derivative of the population state.
  """
  regret = fitness - state.dot(fitness)
  return state * (state * regret - np.sum(state**2 * regret))


class SinglePopulationDynamics(object):
  """Continuous-time single population dynamics.

  Attributes:
    payoff_matrix: The payoff matrix as an `numpy.ndarray` of shape `[2, k_1,
      k_2]`, where `k_1` is the number of strategies of the first player and
      `k_2` for the second player. The game is assumed to be symmetric.
    dynamics: A callback function that returns the time-derivative of the
      population state.
  """

  def __init__(self, payoff_matrix, dynamics):
    """Initializes the single-population dynamics."""
    assert payoff_matrix.ndim == 3
    assert payoff_matrix.shape[0] == 2
    assert np.allclose(payoff_matrix[0], payoff_matrix[1].T)
    self.payoff_matrix = payoff_matrix[0]
    self.dynamics = dynamics

  def __call__(self, state=None, time=None):
    """Time derivative of the population state.

    Args:
      state: Probability distribution as list or
        `numpy.ndarray(shape=num_strategies)`.
      time: Time is ignored (time-invariant dynamics). Including the argument in
        the function signature supports numerical integration via e.g.
        `scipy.integrate.odeint` which requires that the callback function has
        at least two arguments (state and time).

    Returns:
      Time derivative of the population state as
      `numpy.ndarray(shape=num_strategies)`.
    """
    state = np.array(state)
    assert state.ndim == 1
    assert state.shape[0] == self.payoff_matrix.shape[0]
    # (Ax')' = xA'
    fitness = np.matmul(state, self.payoff_matrix.T)
    return self.dynamics(state, fitness)


class MultiPopulationDynamics(object):
  """Continuous-time multi-population dynamics.

  Attributes:
    payoff_tensor: The payoff tensor as an numpy.ndarray of size `[n, k0, k1,
      k2, ...]`, where n is the number of players and `k0` is the number of
      strategies of the first player, `k1` of the second player and so forth.
    dynamics: List of callback functions for the time-derivative of the
      population states, where `dynamics[i]` computes the time-derivative of the
      i-th player's population state. If at construction, only a single callback
      function is provided, the same function is used for all populations.
  """

  def __init__(self, payoff_tensor, dynamics):
    """Initializes the multi-population dynamics."""
    if isinstance(dynamics, list) or isinstance(dynamics, tuple):
      assert payoff_tensor.shape[0] == len(dynamics)
    else:
      dynamics = [dynamics] * payoff_tensor.shape[0]
    self.payoff_tensor = payoff_tensor
    self.dynamics = dynamics

  def __call__(self, state, time=None):
    """Time derivative of the population states.

    Args:
      state: Combined population state for all populations as a list or flat
        `numpy.ndarray` (ndim=1). Probability distributions are concatenated in
        order of the players.
      time: Time is ignored (time-invariant dynamics). Including the argument in
        the function signature supports numerical integration via e.g.
        `scipy.integrate.odeint` which requires that the callback function has
        at least two arguments (state and time).

    Returns:
      Time derivative of the combined population state as `numpy.ndarray`.
    """
    state = np.array(state)
    n = self.payoff_tensor.shape[0]  # number of players
    ks = self.payoff_tensor.shape[1:]  # number of strategies for each player
    assert state.shape[0] == sum(ks)

    states = np.split(state, np.cumsum(ks)[:-1])
    dstates = [None] * n
    for i in range(n):
      # move i-th population to front
      fitness = np.moveaxis(self.payoff_tensor[i], i, 0)
      # marginalize out all other populations
      for i_ in set(range(n)) - {i}:
        fitness = np.tensordot(states[i_], fitness, axes=[0, 1])
      dstates[i] = self.dynamics[i](states[i], fitness)

    return np.concatenate(dstates)


def time_average(traj):
  """Time-averaged population state trajectory.

  Args:
    traj: Trajectory as `numpy.ndarray`. Time is along the first dimension,
      types/strategies along the second.

  Returns:
    Time-averaged trajectory.
  """
  n = traj.shape[0]
  sum_traj = np.cumsum(traj, axis=0)
  norm = 1. / np.arange(1, n + 1)
  return sum_traj * norm[:, np.newaxis]



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for open_spiel.python.algorithms.double_oracle."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import double_oracle
import pyspiel


class DoubleOracleTest(absltest.TestCase):

  def test_rock_paper_scissors(self):
    game = pyspiel.load_matrix_game("matrix_rps")
    solver = double_oracle.DoubleOracleSolver(game)
    solution, iteration, value = solver.solve(initial_strategies=[[0], [0]])
    np.testing.assert_allclose(solution[0], np.ones(3)/3.)
    np.testing.assert_allclose(solution[1], np.ones(3)/3.)
    self.assertEqual(iteration, 3)
    self.assertAlmostEqual(value, 0.0)

  def test_single_step(self):
    game = pyspiel.load_matrix_game("matrix_rps")
    solver = double_oracle.DoubleOracleSolver(game)
    solver.subgame_strategies = [[0], [0]]
    best_response, best_response_utility = solver.step()
    self.assertListEqual(best_response, [1, 1])
    self.assertListEqual(best_response_utility, [1.0, 1.0])

  def test_kuhn_poker(self):
    game = pyspiel.extensive_to_matrix_game(pyspiel.load_game("kuhn_poker"))
    solver = double_oracle.DoubleOracleSolver(game)
    solution, iteration, value = solver.solve(initial_strategies=[[0], [0]])

    # check if solution is Nash
    exp_utilty = solution[0] @ solver.payoffs @ solution[1]
    self.assertAlmostEqual(max(solver.payoffs[0] @ solution[1]), exp_utilty[0])
    self.assertAlmostEqual(max(solution[0] @ solver.payoffs[1]), exp_utilty[1])

    self.assertEqual(iteration, 8)
    self.assertAlmostEqual(value, 0.0)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example algorithm to sample some states from a game."""

import random


def sample_some_states(game, max_states=100):
  """Samples some states in the game.

  This can be run for large games, in contrast to `get_all_states`. It is useful
  for tests that need to check a predicate only on a subset of the game, since
  generating the whole game is infeasible.

  Currently only works for sequential games.

  The algorithm maintains a list of states and repeatedly picks a random state
  from the list to expand until enough states have been sampled.

  Arguments:
    game: The game to analyze, as returned by `load_game`.
    max_states: The maximum number of states to return. Negative means no limit.

  Returns:
    A `list` of `pyspiel.State`.
  """
  states = []
  unexplored_actions = []
  indexes_with_unexplored_actions = set()

  def add_state(state):
    states.append(state)
    if state.is_terminal():
      unexplored_actions.append(None)
    else:
      indexes_with_unexplored_actions.add(len(states) - 1)
      unexplored_actions.append(set(state.legal_actions()))

  def expand_random_state():
    index = random.choice(list(indexes_with_unexplored_actions))
    state = states[index]
    actions = unexplored_actions[index]
    action = random.choice(list(actions))
    actions.remove(action)
    if not actions:
      indexes_with_unexplored_actions.remove(index)
    return state.child(action)

  add_state(game.new_initial_state())
  while (len(states) < max_states) and indexes_with_unexplored_actions:
    add_state(expand_random_state())

  if not states:
    raise ValueError("get_some_states sampled 0 states!")

  return states



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for open_spiel.python.algorithms.masked_softmax."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python.algorithms import masked_softmax

# Temporarily disable TF2 behavior until the code is updated.
tf.disable_v2_behavior()


exp = math.exp  # For shorter lines

_BATCH_INPUTS = np.asarray([
    [1.0, 2.0, 3.0],
    [4.0, 5.0, 6.0],
    [7.0, 8.0, 9.0],
    [10.0, 11.0, 12.0],
    [13.0, 14.0, 15.0],
    [16.0, 17.0, 18.0],
])
_BATCH_MASK = np.asarray([
    [1.0, 1.0, 1.0],
    [1.0, 0.0, 1.0],
    [0.0, 1.0, 1.0],
    [0.0, 0.0, 1.0],
    [0.0, 1.0, 0.0],
    [1.0, 0.0, 0.0],
])
total_row_0 = exp(1) + exp(2) + exp(3)
total_row_1 = exp(4) + exp(6)
total_row_2 = exp(8) + exp(9)
# pyformat: disable
_BATCH_EXPECTED = np.asarray([
    [exp(1) / total_row_0, exp(2) / total_row_0, exp(3) / total_row_0],
    [exp(4) / total_row_1, 0, exp(6) / total_row_1],
    [0, exp(8) / total_row_2, exp(9) / total_row_2],
    [0, 0, 1],
    [0, 1, 0],
    [1, 0, 0],
])
# pyformat: enable

# The following provides a 2-batch set of time-sequence policies.
# [B, T, num_actions] = 2, 3, 3
_B_T_LOGITS = np.asarray([[
    [1.0, 2.0, 3.0],
    [4.0, 5.0, 6.0],
    [7.0, 8.0, 9.0],
], [
    [10.0, 11.0, 12.0],
    [13.0, 14.0, 15.0],
    [16.0, 17.0, 18.0],
]])
_B_T_MASK = np.asarray([[
    [1.0, 1.0, 1.0],
    [1.0, 0.0, 1.0],
    [0.0, 1.0, 1.0],
], [
    [0.0, 0.0, 1.0],
    [0.0, 1.0, 0.0],
    [1.0, 0.0, 0.0],
]])
_B_T_EXPECTED = np.asarray([[
    [exp(1) / total_row_0,
     exp(2) / total_row_0,
     exp(3) / total_row_0],
    [exp(4) / total_row_1, 0, exp(6) / total_row_1],
    [0, exp(8) / total_row_2, exp(9) / total_row_2],
], [
    [0, 0, 1],
    [0, 1, 0],
    [1, 0, 0],
]])
array = np.asarray
# We test over all the above examples.
_ALL_TESTS_INPUTS = [
    # Non-batch inputs
    (array([1., 1.]), array([1., 1.]), array([.5, .5])),
    (array([1., 1.]), array([0., 1.]), array([0., 1.])),
    (array([1., 1.]), array([1., 0.]), array([1., 0.])),
    (array([1., 1., 1]), array([1., 1., 0]), array([.5, .5, 0.])),
    # Batch-inputs
    (_BATCH_INPUTS, _BATCH_MASK, _BATCH_EXPECTED),
    # Batch-time inputs
    (_B_T_LOGITS, _B_T_MASK, _B_T_EXPECTED),
]


class MaskedSoftmaxTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(_ALL_TESTS_INPUTS)
  def test_np_masked_softmax(self, logits, legal_actions, expected):
    np.testing.assert_array_almost_equal(
        expected, masked_softmax.np_masked_softmax(logits, legal_actions))

  @parameterized.parameters(_ALL_TESTS_INPUTS)
  def test_tf_masked_softmax(self, np_logits, np_legal_actions, expected):
    logits = tf.Variable(np_logits, tf.float32)
    mask = tf.Variable(np_legal_actions, tf.float32)

    policy = masked_softmax.tf_masked_softmax(logits, mask)

    with tf.Session() as sess:
      sess.run(tf.global_variables_initializer())
      np_policy = sess.run(policy)

    np.testing.assert_array_almost_equal(expected, np_policy)

  def test_masked_softmax_on_all_invalid_moves(self):
    # If all actions are illegal, the behavior is undefined (it can be nan
    # or can be 0. We add this test to document this behavior and know if we
    # change it.
    np_logits = np.asarray([[
        [1.0, 1.0, 1.0],
        [0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0],
    ]])
    logits = tf.Variable(np_logits, tf.float32)
    np_mask = np.asarray([[
        [1.0, 1.0, 1.0],
        [1.0, 0.0, 1.0],
        [0.0, 0.0, 0.0],
    ]])
    mask = tf.Variable(np_mask, tf.float32)

    expected = np.asarray([[
        [1 / 3, 1 / 3, 1 / 3],
        [1 / 2, 0.0, 1 / 2],
        [np.nan, np.nan, np.nan],
    ]])

    policy = masked_softmax.tf_masked_softmax(logits, mask)

    with tf.Session() as sess:
      sess.run(tf.global_variables_initializer())
      np_policy = sess.run(policy)
    np.testing.assert_array_almost_equal(expected, np_policy)

    # Numpy behaves similarly.
    np.testing.assert_array_almost_equal(
        expected, masked_softmax.np_masked_softmax(np_logits, np_mask))


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exploitability of a policy from IS-MCTS search run at each info state."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")

SEED = 129846127


def construct_is_mcts_policy(game, state, tabular_policy, bot, searched):
  """Constructs a tabular policy from independent bot calls.

  Args:
    game: an OpenSpiel game,
    state: an OpenSpiel state to start the tree walk from,
    tabular_policy: a policy.TabularPolicy for this game,
    bot: the bot to get the policy from at each state
    searched: a dictionary of information states already search (empty to begin)
  """

  if state.is_terminal():
    return
  elif state.is_chance_node():
    outcomes = state.legal_actions()
    for outcome in outcomes:
      new_state = state.clone()
      new_state.apply_action(outcome)
      construct_is_mcts_policy(game, new_state, tabular_policy, bot, searched)
  else:
    infostate_key = state.information_state_string()
    if infostate_key not in searched:
      searched[infostate_key] = True
      infostate_policy = bot.get_policy(state)
      tabular_state_policy = tabular_policy.policy_for_key(infostate_key)
      for action, prob in infostate_policy:
        tabular_state_policy[action] = prob
    for action in state.legal_actions():
      new_state = state.clone()
      new_state.apply_action(action)
      construct_is_mcts_policy(game, new_state, tabular_policy, bot, searched)


def main(_):
  game = pyspiel.load_game(FLAGS.game)
  evaluator = pyspiel.RandomRolloutEvaluator(1, SEED)
  min_expl = game.max_utility() -  game.min_utility()

  print("{:>5} {:>10} {:>50} {:>20}".format(
      "max_sims", "uct_c", "final_policy_type", "exploitability"))
  for max_simulations in [10, 100, 1000, 10000]:
    for uct_c in [0.2, 0.5, 1.0, 2.0, 4.0]:  # These values are for Kuhn.
      for final_policy_type in [
          pyspiel.ISMCTSFinalPolicyType.NORMALIZED_VISIT_COUNT,
          pyspiel.ISMCTSFinalPolicyType.MAX_VISIT_COUNT,
          pyspiel.ISMCTSFinalPolicyType.MAX_VALUE
      ]:
        tabular_policy = policy.TabularPolicy(game)
        bot = pyspiel.ISMCTSBot(SEED, evaluator, uct_c, max_simulations, -1,
                                final_policy_type, False, False)
        searched = {}
        construct_is_mcts_policy(game, game.new_initial_state(), tabular_policy,
                                 bot, searched)
        expl = exploitability.exploitability(game, tabular_policy)
        print("{:>5} {:>10} {:>50} {:>20}".format(max_simulations, uct_c,
                                                  str(final_policy_type), expl))
        if expl < min_expl:
          min_expl = expl
  print("Min expl: {}".format(min_expl))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python implementation for Monte Carlo Counterfactual Regret Minimization."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum
import numpy as np
import pyspiel

# Indices in the information sets for the regrets and average policy sums.
_REGRET_INDEX = 0
_AVG_POLICY_INDEX = 1


class AverageType(enum.Enum):
  SIMPLE = 0
  FULL = 1


class ExternalSamplingSolver(object):
  """An implementation of external sampling MCCFR."""

  def __init__(self, game, average_type=AverageType.SIMPLE):
    self._game = game
    self._infostates = {}  # infostate keys -> [regrets, avg strat]
    self._num_players = game.num_players()
    # How to average the strategy. The 'simple' type does the averaging for
    # player i + 1 mod num_players on player i's regret update pass; in two
    # players this corresponds to the standard implementation (updating the
    # average policy at opponent nodes). In n>2 players, this can be a problem
    # for several reasons: first, it does not compute the estimate as described
    # by the (unbiased) stochastically-weighted averaging in chapter 4 of
    # Lanctot 2013  commonly used in MCCFR because the denominator (important
    # sampling correction) should include all the other sampled players as well
    # so the sample reach no longer cancels with reach of the player updating
    # their average policy. Second, if one player assigns zero probability to an
    # action (leading to a subtree), the average policy of a different player in
    # that subtree is no longer updated. Hence, the full averaging does not
    # update the average policy in the regret passes but does a separate pass to
    # update the average policy. Nevertheless, we set the simple type as the
    # default because it is faster, seems to work better empirically, and it
    # matches what was done in Pluribus (Brown and Sandholm. Superhuman AI for
    # multiplayer poker. Science, 11, 2019).
    self._average_type = average_type

    assert game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL, (
        "MCCFR requires sequential games. If you're trying to run it " +
        'on a simultaneous (or normal-form) game, please first transform it ' +
        'using turn_based_simultaneous_game.')

  def iteration(self):
    """Performs one iteration of external sampling.

    An iteration consists of one episode for each player as the update
    player.
    """
    for player in range(self._num_players):
      self._update_regrets(self._game.new_initial_state(), player)
    if self._average_type == AverageType.FULL:
      reach_probs = np.ones(self._num_players, dtype=np.float64)
      self._full_update_average(self._game.new_initial_state(), reach_probs)

  def _lookup_infostate_info(self, info_state_key, num_legal_actions):
    """Looks up an information set table for the given key.

    Args:
      info_state_key: information state key (string identifier).
      num_legal_actions: number of legal actions at this information state.

    Returns:
      A list of:
        - the average regrets as a numpy array of shape [num_legal_actions]
        - the average strategy as a numpy array of shape
        [num_legal_actions].
          The average is weighted using `my_reach`
    """
    retrieved_infostate = self._infostates.get(info_state_key, None)
    if retrieved_infostate is not None:
      return retrieved_infostate

    # Start with a small amount of regret and total accumulation, to give a
    # uniform policy: this will get erased fast.
    self._infostates[info_state_key] = [
        np.ones(num_legal_actions, dtype=np.float64) / 1000.0,
        np.ones(num_legal_actions, dtype=np.float64) / 1000.0,
    ]
    return self._infostates[info_state_key]

  def _add_regret(self, info_state_key, action_idx, amount):
    self._infostates[info_state_key][_REGRET_INDEX][action_idx] += amount

  def _add_avstrat(self, info_state_key, action_idx, amount):
    self._infostates[info_state_key][_AVG_POLICY_INDEX][action_idx] += amount

  def callable_avg_policy(self):
    """Returns the average joint policy as a callable.

    The callable has a signature of the form string (information
    state key) -> list of (action, prob).
    """

    def wrap(state):
      info_state_key = state.information_state_string(state.current_player())
      legal_actions = state.legal_actions()
      infostate_info = self._lookup_infostate_info(info_state_key,
                                                   len(legal_actions))
      avstrat = (
          infostate_info[_AVG_POLICY_INDEX] /
          infostate_info[_AVG_POLICY_INDEX].sum())
      return [(legal_actions[i], avstrat[i]) for i in range(len(legal_actions))]

    return wrap

  def _regret_matching(self, regrets, num_legal_actions):
    """Applies regret matching to get a policy.

    Args:
      regrets: numpy array of regrets for each action.
      num_legal_actions: number of legal actions at this state.

    Returns:
      numpy array of the policy indexed by the index of legal action in the
      list.
    """
    positive_regrets = np.maximum(regrets,
                                  np.zeros(num_legal_actions, dtype=np.float64))
    sum_pos_regret = positive_regrets.sum()
    if sum_pos_regret <= 0:
      return np.ones(num_legal_actions, dtype=np.float64) / num_legal_actions
    else:
      return positive_regrets / sum_pos_regret

  def _full_update_average(self, state, reach_probs):
    """Performs a full update average.

    Args:
      state: the open spiel state to run from
      reach_probs: array containing the probability of reaching the state
        from the players point of view
    """
    if state.is_terminal():
      return
    if state.is_chance_node():
      for action in state.legal_actions():
        self._full_update_average(state.child(action), reach_probs)
      return

    # If all the probs are zero, no need to keep going.
    sum_reach_probs = np.sum(reach_probs)
    if sum_reach_probs == 0:
      return

    cur_player = state.current_player()
    info_state_key = state.information_state_string(cur_player)
    legal_actions = state.legal_actions()
    num_legal_actions = len(legal_actions)

    infostate_info = self._lookup_infostate_info(info_state_key,
                                                 num_legal_actions)
    policy = self._regret_matching(infostate_info[_REGRET_INDEX],
                                   num_legal_actions)

    for action_idx in range(num_legal_actions):
      new_reach_probs = np.copy(reach_probs)
      new_reach_probs[cur_player] *= policy[action_idx]
      self._full_update_average(
          state.child(legal_actions[action_idx]), new_reach_probs)

    # Now update the cumulative policy
    for action_idx in range(num_legal_actions):
      self._add_avstrat(info_state_key, action_idx,
                        reach_probs[cur_player] * policy[action_idx])

  def _update_regrets(self, state, player):
    """Runs an episode of external sampling.

    Args:
      state: the open spiel state to run from
      player: the player to update regrets for

    Returns:
      value: is the value of the state in the game
      obtained as the weighted average of the values
      of the children
    """
    if state.is_terminal():
      return state.player_return(player)

    if state.is_chance_node():
      outcomes, probs = zip(*state.chance_outcomes())
      outcome = np.random.choice(outcomes, p=probs)
      return self._update_regrets(state.child(outcome), player)

    cur_player = state.current_player()
    info_state_key = state.information_state_string(cur_player)
    legal_actions = state.legal_actions()
    num_legal_actions = len(legal_actions)

    infostate_info = self._lookup_infostate_info(info_state_key,
                                                 num_legal_actions)
    policy = self._regret_matching(infostate_info[_REGRET_INDEX],
                                   num_legal_actions)

    value = 0
    child_values = np.zeros(num_legal_actions, dtype=np.float64)
    if cur_player != player:
      # Sample at opponent node
      action_idx = np.random.choice(np.arange(num_legal_actions), p=policy)
      value = self._update_regrets(
          state.child(legal_actions[action_idx]), player)
    else:
      # Walk over all actions at my node
      for action_idx in range(num_legal_actions):
        child_values[action_idx] = self._update_regrets(
            state.child(legal_actions[action_idx]), player)
        value += policy[action_idx] * child_values[action_idx]

    if cur_player == player:
      # Update regrets.
      for action_idx in range(num_legal_actions):
        self._add_regret(info_state_key, action_idx,
                         child_values[action_idx] - value)
    # Simple average does averaging on the opponent node. To do this in a game
    # with more than two players, we only update the player + 1 mod num_players,
    # which reduces to the standard rule in 2 players.
    if self._average_type == AverageType.SIMPLE and cur_player == (
        player + 1) % self._num_players:
      for action_idx in range(num_legal_actions):
        self._add_avstrat(info_state_key, action_idx, policy[action_idx])

    return value



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A class for storing iteration-specific metrics.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class IterationStatistics(object):
  """A class for storing iteration-specific metrics.

  The internal format is as follows: we maintain a mapping from keys to lists.
  Each list contains all the values corresponding to the given key.

  For example, self.data_lists['train_episode_returns'] might contain the
    per-episode returns achieved during this iteration.

  Attributes:
    data_lists: dict mapping each metric_name (str) to a list of said metric
      across episodes.
  """

  def __init__(self):
    self.data_lists = {}

  def append(self, data_pairs):
    """Add the given values to their corresponding key-indexed lists.

    Args:
      data_pairs: A dictionary of key-value pairs to be recorded.
    """
    for key, value in data_pairs.items():
      if key not in self.data_lists:
        self.data_lists[key] = []
      self.data_lists[key].append(value)
